# -*- coding: utf-8 -*-

from typing import Optional, List
from copy import deepcopy

from yaml import YAMLError

from PyQt5 import uic
from PyQt5.QtGui import QRegularExpressionValidator
from PyQt5.QtCore import QProcess, Qt, QCoreApplication, QRegularExpression, QPoint
from PyQt5.QtWidgets import QDialog, QWidget, QDialogButtonBox, QListWidget, QPushButton, QGroupBox, \
    QSpinBox, QListWidgetItem, QAbstractButton, QLabel, QLineEdit, QToolTip

from libastra_csp.yaml_settings import YamlSettings
from libastra_csp.utils import Utils
from libastra_csp.common_settings import CommonSettings
from libfly_csp.graphic_settings import GraphicSettings

translate = QCoreApplication.translate


class SettingsDialog(QDialog):
    def __init__(self,
                 common_settings: CommonSettings,
                 graphic_settings: GraphicSettings,
                 parent: QWidget = None) -> None:
        super(SettingsDialog, self).__init__(parent=parent)
        self.buttonBox = None  # type: Optional[QDialogButtonBox]
        self.enabledExtensionsList = None  # type: Optional[QListWidget]
        self.maxCspTimeWaiting = None  # type: Optional[QSpinBox]
        self.maxMacLevel = None  # type: Optional[QSpinBox]
        self.pdfConversionList = None  # type: Optional[QListWidget]
        self.extensionFilterGroupbox = None  # type: Optional[QGroupBox]
        self.addExtensionButton = None  # type: Optional[QPushButton]
        self.removeExtensionButton = None  # type: Optional[QPushButton]
        self.addMimeTypeButton = None  # type: Optional[QPushButton]
        self.removeMimeTypeButton = None  # type: Optional[QPushButton]
        uic.loadUi(Utils.file_path(__file__, 'ui', 'settings.ui'), self)

        self.__process = None  # type: Optional[QProcess]
        self.__pressed_button = None  # type: Optional[QDialogButtonBox.StandardButton]
        self.__common_settings = common_settings
        self.__graphic_settings = graphic_settings

        self.__setup()

    def __setup(self) -> None:
        self.setModal(True)
        self.adjustSize()
        self.maxMacLevel.setMinimum(0)
        self.maxCspTimeWaiting.setMinimum(-1)
        self.maxCspTimeWaiting.setMaximum(99999)
        self.removeExtensionButton.clicked.connect(self.__on_remove_extension_clicked)
        self.removeMimeTypeButton.clicked.connect(self.__on_remove_mime_type_clicked)
        self.addExtensionButton.clicked.connect(self.__on_add_extension_clicked)
        self.addMimeTypeButton.clicked.connect(self.__on_add_mime_type_clicked)
        self.buttonBox.clicked.connect(self.__on_button_box_clicked)

    def __on_remove_extension_clicked(self) -> None:
        self.enabledExtensionsList.takeItem(self.enabledExtensionsList.currentRow())

    def __on_remove_mime_type_clicked(self) -> None:
        self.pdfConversionList.takeItem(self.pdfConversionList.currentRow())

    def __on_add_extension_clicked(self) -> None:
        self.__input_dialog(translate('text', 'Add extension'),
                            translate('text', 'Type extension:'),
                            self.enabledExtensionsList,
                            self.__common_settings.extension_regex)

    def __on_add_mime_type_clicked(self) -> None:
        self.__input_dialog(translate('text', 'Add MIME type'),
                            translate('text', 'Type MIME type:'),
                            self.pdfConversionList,
                            self.__graphic_settings.mime_types_regex)

    def __input_dialog(self, label: str, text: str, list_widget: QListWidget, validator_string: str) -> None:
        dialog = InputDialog(self, label, text, list_widget, validator_string)
        dialog.open()

    def __on_button_box_clicked(self, button: QAbstractButton) -> None:
        standard_button = self.buttonBox.standardButton(button)
        if standard_button == QDialogButtonBox.Cancel:
            self.reject()
            return
        if self.__pressed_button == QDialogButtonBox.Apply or \
                self.__pressed_button == QDialogButtonBox.Ok:
            return
        self.__pressed_button = standard_button
        self.__save()

    def __save(self) -> None:
        process = QProcess()
        process.setProgram('pkexec')
        try:
            arguments = ['/usr/libexec/flycspconfigwriter',
                         '--common-settings={data}'.format(data=self.__common_settings_data()),
                         '--graphic-settings={data}'.format(data=self.__graphic_settings_data())]
        except YAMLError as exception:
            self.__pressed_button = None
            self.parent().show_error(translate('text', 'Settings error'),
                                     translate('text', 'Can\'t save settings.'),
                                     str(exception))
            return
        process.setArguments(arguments)
        process.finished.connect(self.__on_save_process_finished)
        self.__process = process
        self.__process.start()

    def __on_save_process_finished(self, exit_code: int, exit_status: QProcess.ExitStatus) -> None:
        if exit_status != QProcess.NormalExit or exit_code != 0:
            self.parent().show_error(translate('text', 'Process crashed'),
                                     Utils.to_utf8(self.__process.readAllStandardError()),
                                     translate('text', 'Exit code: {code}'.format(code=exit_code)))
        else:
            self.__common_settings.load()
            self.__graphic_settings.load()
            self.__pressed_button != QDialogButtonBox.Ok or self.accept()
        self.__pressed_button = None

    def __common_settings_data(self) -> str:
        settings = deepcopy(self.__common_settings)
        settings.max_mac_level = self.maxMacLevel.value()
        settings.max_csp_time_waiting = self.maxCspTimeWaiting.value()
        settings.file_extensions_filter = {'enable': self.extensionFilterGroupbox.isChecked(),
                                           'extensions': self.__get_data_from_widget(self.enabledExtensionsList)}
        return self.__serialized_settings(settings)

    def __graphic_settings_data(self) -> str:
        settings = deepcopy(self.__graphic_settings)
        settings.pdf_conversion_mime_types = self.__get_data_from_widget(self.pdfConversionList)
        return self.__serialized_settings(settings)

    @staticmethod
    def __serialized_settings(settings: YamlSettings) -> str:
        serialized = settings.serialize()
        if settings.error:
            raise YAMLError(settings.error)
        return serialized

    @staticmethod
    def __get_data_from_widget(list_widget: QListWidget) -> List[str]:
        data = []
        for row in range(list_widget.count()):
            data.append(list_widget.item(row).text().strip())
        return data

    @staticmethod
    def __insert_items_to_list(items: List[str], list_widget: QListWidget) -> None:
        for item in items:
            list_widget.addItem(create_item(item))

    def emplace_data(self) -> None:
        self.pdfConversionList.clear()
        self.enabledExtensionsList.clear()

        self.maxMacLevel.setValue(self.__common_settings.max_mac_level)
        self.maxCspTimeWaiting.setValue(self.__common_settings.max_csp_time_waiting)
        self.extensionFilterGroupbox.setChecked(self.__common_settings.file_extensions_filter.get('enable', False))
        self.__insert_items_to_list(items=self.__common_settings.file_extensions_filter.get('extensions', []),
                                    list_widget=self.enabledExtensionsList)
        self.__insert_items_to_list(items=self.__graphic_settings.pdf_conversion_mime_types or [],
                                    list_widget=self.pdfConversionList)


def create_item(text: str) -> QListWidgetItem:
    item = QListWidgetItem()
    item.setText(text)
    item.setFlags(item.flags() | Qt.ItemIsEditable)
    return item

class InputDialog(QDialog):
    def __init__(self, parent: QWidget, label: str, title: str, list_widget: QListWidget,
                 validator_string: str) -> None:
        super(InputDialog, self).__init__(parent=parent)
        self.label = None # type: Optional[QLabel]
        self.lineEdit = None # type: Optional[QLineEdit]
        self.buttonBox = None # type: Optional[QDialogButtonBox]
        self.list_widget = list_widget
        uic.loadUi(Utils.file_path(__file__, 'ui', 'inputdialog.ui'), self)

        self.lineEdit.setValidator(QRegularExpressionValidator(QRegularExpression(validator_string)))

        self.buttonBox.clicked.connect(self.__on_button_box_clicked)

        self.setWindowTitle(title)
        self.label.setText(label)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

    def __on_button_box_clicked(self, button: QAbstractButton) -> None:
        standard_button = self.buttonBox.standardButton(button)
        if standard_button == QDialogButtonBox.Cancel:
            self.done(QDialog.Rejected)
        elif standard_button == QDialogButtonBox.Ok:
            if self.lineEdit.hasAcceptableInput() :
               if not self.__has_duplicate_item(self.list_widget, self.lineEdit.text()):
                    self.list_widget.addItem(create_item(self.lineEdit.text()))
                    self.done(QDialog.Accepted)
               else:
                   self.__error_tooltip(translate('text', 'Item is on the list already'))
            else:
                self.__error_tooltip(translate('text','Invalid input'))

    def __error_tooltip(self, text: str):
        lower_bound_line_edit = self.lineEdit.height() - 10
        QToolTip.showText(self.lineEdit.mapToGlobal(QPoint(0, lower_bound_line_edit)), text)

    @staticmethod
    def __has_duplicate_item(list_widget: QListWidget, item: str) -> bool:
        items = list_widget.findItems(item, Qt.MatchExactly)
        return item in [item.text() for item in items]
    